/**
 * @file srcMLReader.cpp
 * @copyright
 *
 * Copyright (C) 2013-2014  SDML (www.srcML.org)
 *
 * The srcML Toolkit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The srcML Toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the srcML Toolkit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <srcMLReader.hpp>
#include <srcmlns.hpp>
#include <srcml.h>

#include <stdlib.h>
#include <string.h>

#include <sstream>

void output_node_srcml(const xmlNode & node, xmlTextWriterPtr writer);

/**
 * getNode
 * @param reader an XML text reader
 *
 * @returns Get a copy of the current node with attributes set.
 */
xmlNodePtr getNode(xmlTextReaderPtr reader) {

  xmlNodePtr node = xmlCopyNode(xmlTextReaderCurrentNode(reader), 2);
  node->extra = xmlTextReaderIsEmptyElement(reader);
  node->type = (xmlElementType)xmlTextReaderNodeType(reader);

  return node;
}

/**
 * freeNode
 * @param node an XML node
 *
 * Free the current node.  Text is not allocated and do not free
 */
void freeNode(xmlNodePtr & node) {

  if(node) {

    if(node->type == XML_ELEMENT_DECL) node->type = XML_ELEMENT_NODE;

    if((xmlReaderTypes)node->type != XML_READER_TYPE_TEXT
       && (xmlReaderTypes)node->type != XML_READER_TYPE_SIGNIFICANT_WHITESPACE)
      xmlFreeNode(node);
    else {

      if(node->content) xmlFree(node->content);
      xmlFree(node);

    }

  }

  node = 0;

}


/**
 * srcMLReader
 * @param filename name of a file
 *
 * Construct a srcMLReader using a filename
 */
srcMLReader::srcMLReader(const char * filename)
  : is_archive(false), is_single(false), done(false) {

  if(filename == NULL) throw std::string();

  reader = xmlNewTextReaderFilename(filename);
  if(reader == NULL) throw std::string();
  xmlTextReaderRead(reader);
  node = getNode(reader);
}

/**
 * srcMLReader
 * @param input parser input buffer
 *
 * Construct a srcMLReader using a parser input buffer
 */
srcMLReader::srcMLReader(xmlParserInputBufferPtr input)
  : is_archive(false), is_single(false), done(false) {

  if(input == NULL) throw std::string();
  reader = xmlNewTextReader(input, 0);
  if(reader == NULL) throw std::string();
  xmlTextReaderRead(reader);
  node = getNode(reader);
}

/**
 * ~srcMLReader
 *
 * Destructor a srcMLReader
 */
srcMLReader::~srcMLReader() {

  if(!save_nodes.empty()) {

    try {

      for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size(); ++i)
        freeNode(save_nodes.at(i));

    } catch(...) {}

    save_nodes.clear();

  }

  freeNode(node);
  xmlFreeTextReader(reader);
  reader = 0;

}

/**
 * readAttributesInternal
 * @param language a location to store the language attribute
 * @param filename a location to store the filename attribute
 * @param directory a location to store the directory attribute
 * @param version a location to store the version attribute
 *
 * Internal function to read unit attributes.
 *
 * @returns 0 on success and 1 on failure.
 */
int srcMLReader::readUnitAttributesInternal(std::string *& language, std::string *& filename,
                                            std::string *& directory, std::string *& version) {

  xmlAttrPtr attribute = save_nodes.back()->properties;
  while (attribute) {

    std::string name = (const char *)attribute->name;
    try {

      if(name == "language")
        language = new std::string((const char *)attribute->children->content);
      else if(name == "filename")
        filename = new std::string((const char *)attribute->children->content);
      else if(name == "dir")
        directory = new std::string((const char *)attribute->children->content);
      else if(name == "version")
        version = new std::string((const char *)attribute->children->content);

    } catch(...) {

      if(language) delete language, language = 0;
      if(filename) delete filename, filename = 0;
      if(directory) delete directory, directory = 0;
      if(version) delete version, version = 0;
      return 1;

    }

    attribute = attribute->next;
  }

  return 0;

}

/**
 * readRootUnitAttributes
 * @param language a location to store the language attribute
 * @param filename a location to store the filename attribute
 * @param directory a location to store the directory attribute
 * @param version a location to store the version attribute
 * @param attributes array to store other attributes gathered
 * @param prefixes an array to store gathered XML namespace prefixes
 * @param namespaces an array to store gathered XML naamespaces
 * @param options a variable to set used options
 * @param tabstop a variable to set the tabstop
 *
 * Read attributes and namespace information fromt the root unit,
 * setting the necessary options.
 *
 * @returns 1 on success and 0 on failure.
 */
int srcMLReader::readRootUnitAttributes(std::string *& language, std::string *& filename,
                                        std::string *& directory, std::string *& version,
                                        std::vector<std::string> & attributes,
                                        std::vector<std::string> & prefixes,
                                        std::vector<std::string> & namespaces,
                                        OPTION_TYPE & options,
                                        int & tabstop) {

  if(done) return 0;
  if(is_single || is_archive) return 0;

  // forward to start unit
  while(true) {
    if(node && (xmlReaderTypes)node->type == XML_READER_TYPE_ELEMENT && strcmp((const char *)node->name, "unit") == 0)
      break;

    if(xmlTextReaderRead(reader) != 1) { done = true; return 0; }
    freeNode(node);
    node = getNode(reader);
  }

  xmlAttrPtr attribute = node->properties;
  while (attribute) {
    std::string name = (const char *)attribute->name;
    std::string value = (const char *)attribute->children->content;

    try {

      if(name == "language")
        language = new std::string(value);
      else if(name == "filename")
        filename = new std::string(value);
      else if(name == "dir")
        directory = new std::string(value);
      else if(name == "version")
        version = new std::string(value);
      else if(name == "tabs")
        tabstop = atoi(value.c_str());
      else if(name == "options") {

	while(!value.empty()) {

	  std::string::size_type pos = value.find(",");
	  std::string option = value.substr(0, pos);
	  if(pos == std::string::npos)
	    value = "";
	  else
	    value = value.substr(value.find(",") + 1);

	  if(option == "XMLDECL")
	    options |= OPTION_XMLDECL;
	  if(option == "NAMESPACEDECL")
	    options |= OPTION_NAMESPACEDECL;
	  if(option == "CPP_TEXT_ELSE")
	    options |= OPTION_CPP_TEXT_ELSE;
	  if(option == "CPP_MARKUP_IF0")
	    options |= OPTION_CPP_MARKUP_IF0;
	  if(option == "EXPRESSION")
	    options |= OPTION_EXPRESSION;
	  if(option == "NAMESPACE")
	    options |= OPTION_NAMESPACE;
	  if(option == "LINE")
	    options |= OPTION_LINE;
	  if(option == "MACRO_PATTERN")
	    options |= OPTION_MACRO_PATTERN;
	  if(option == "MACRO_LIST")
	    options |= OPTION_MACRO_LIST;
	  if(option == "NESTIF")
	    options |= OPTION_NESTIF;
	  if(option == "CPPIF_CHECK")
	    options |= OPTION_CPPIF_CHECK;

	}

      } else {

        attributes.push_back(name);
        attributes.push_back(value);

      }

    } catch(...) {

      if(language) delete language, language = 0;
      if(filename) delete filename, filename = 0;
      if(directory) delete directory, directory = 0;
      if(version) delete version, version = 0;
      return 0;

    }

    attribute = attribute->next;
  }

  xmlNsPtr xmlns = node->nsDef;

  while(xmlns) {

    std::string prefix = xmlns->prefix ? (const char *)xmlns->prefix : "";
    std::string ns = xmlns->href ? (const char *)xmlns->href : "";

    if(ns == SRCML_CPP_NS_URI) {

      if(language) {

        if((*language) == "C++" || (*language) == "C")
          options |= SRCML_OPTION_CPP | SRCML_OPTION_CPP_NOMACRO;
        else if((*language) == "C#")
          options |= SRCML_OPTION_CPP_NOMACRO;
        else
          options |= SRCML_OPTION_CPP;
      }

    } else if(ns == SRCML_ERR_NS_URI)
      options |= SRCML_OPTION_DEBUG;
    else if(ns == SRCML_EXT_LITERAL_NS_URI)
      options |= SRCML_OPTION_LITERAL;
    else if(ns == SRCML_EXT_OPERATOR_NS_URI)
      options |= SRCML_OPTION_OPERATOR;
    else if(ns == SRCML_EXT_MODIFIER_NS_URI)
      options |= SRCML_OPTION_MODIFIER;
    else if(ns == SRCML_EXT_POSITION_NS_URI)
      options |= SRCML_OPTION_POSITION;

    std::vector<std::string>::size_type i;
    try {

      for(i = 0; i < prefixes.size(); ++i)

        if(namespaces.at(i) == ns) {

          prefixes.at(i) = prefix;
          break;
        }

    } catch(...) {}

    if(i == prefixes.size()) {
      prefixes.push_back(prefix);
      namespaces.push_back(ns);
    }

    xmlns = xmlns->next;
  }

  return 1;
}

/**
 * readUnitAttributes
 * @param language a location to store the language attribute
 * @param filename a location to store the filename attribute
 * @param directory a location to store the directory attribute
 * @param version a location to store the version attribute
 *
 * Read attributes from next unit.
 * @todo add handling for empty root element.
 *
 * @returns 1 on success and 0 on failure.
 */
int srcMLReader::readUnitAttributes(std::string *& language, std::string *& filename,
                                    std::string *& directory, std::string *& version) {

  if(!save_nodes.empty()) {

    try {

      for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size(); ++i)
        freeNode(save_nodes.at(i));

    } catch(...) {}

    save_nodes.clear();

  }

  if(done) return 0;

  // forward to start unit
  while(true) {
    if(node && (xmlReaderTypes)node->type == XML_READER_TYPE_ELEMENT && strcmp((const char *)node->name, "unit") == 0)
      break;

    if(xmlTextReaderRead(reader) != 1) { done = true; return 0; }
    freeNode(node);
    node = getNode(reader);
  }

  save_nodes.push_back(node);

  if(readUnitAttributesInternal(language, filename, directory, version)) return 0;
  node = 0;

  if(xmlTextReaderRead(reader) != 1) { done = true; if(save_nodes.back()->extra) return 1; return 0; }

  node = getNode(reader);

  if(is_archive) return 1;

  while(true) {

    save_nodes.push_back(node);
    node = 0;

    if(save_nodes.back()->type == (xmlElementType)XML_READER_TYPE_ELEMENT && strcmp((const char *)save_nodes.back()->name, "unit") == 0) {

      is_archive = true;
      try {
        for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size() - 1; ++i)
          freeNode(save_nodes.at(i));

      } catch(...) {}
      xmlNodePtr save_node = save_nodes.back();
      save_nodes.clear();
      save_nodes.push_back(save_node);
      if(language) delete language, language = 0;
      if(filename) delete filename, filename = 0;
      if(directory) delete directory, directory = 0;
      if(version) delete version, version = 0;
      if(readUnitAttributesInternal(language, filename, directory, version)) return 0;
      break;
    }

    if(save_nodes.back()->type == (xmlElementType)XML_READER_TYPE_END_ELEMENT
       && strcmp((const char *)save_nodes.back()->name, "unit") == 0) {
      is_archive = true;
      is_single = true;
      break;
    }

    if(save_nodes.back()->type == (xmlElementType)XML_READER_TYPE_ELEMENT
       && strcmp((const char *)save_nodes.back()->name, "unit") != 0) {
      is_archive = true;
      is_single = true;
      break;
    }

    node = 0;
    if(xmlTextReaderRead(reader) != 1) {done = true; return 0; }
    node = getNode(reader);

  }

  if(xmlTextReaderRead(reader) != 1) { done = true; return 0; }
  node = getNode(reader);

  return 1;

}

/**
 * readsrcML
 * @param writer an XML text writer
 *
 * Read the next unit of a srcML Archive.
 * and write it to the writer.
 * @todo add handling for empty root element.
 *
 * @returns Return 0 when finished and 1 otherwize.
 */
int srcMLReader::readsrcML(xmlTextWriterPtr writer) {

  //xmlTextWriterStartDocument(writer, XML_VERSION, xml_encoding, XML_DECLARATION_STANDALONE);
  bool read_unit_start = false;

  if(!save_nodes.empty()) {

    bool is_empty = false;
    if(strcmp((const char *)save_nodes.back()->name, "unit") == 0 && save_nodes.back()->extra)
      is_empty = true;

    try {

      for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size(); ++i)
        output_node_srcml(*save_nodes.at(i), writer);

      for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size(); ++i)
        freeNode(save_nodes.at(i));

    } catch(...) {}

    save_nodes.clear();
    if(done || is_empty) { xmlTextWriterEndDocument(writer); return 1; }

  } else {

    if(done) return 0;

    // forward to start unit
    while(true) {
      if(node && (xmlReaderTypes)node->type == XML_READER_TYPE_ELEMENT && strcmp((const char *)node->name, "unit") == 0)
        break;

      if(xmlTextReaderRead(reader) != 1) {done = true; return 0; }
      freeNode(node);
      node = getNode(reader);
    }

  }

  while(true) {

    if(is_archive) output_node_srcml(*node, writer);
    else save_nodes.push_back(node);

    if(strcmp((const char *)node->name, "unit") == 0) {

      if(node->type == (xmlElementType)XML_READER_TYPE_ELEMENT) {

        if(read_unit_start || node->extra) {

          try {

            for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size() - 1; ++i)
              freeNode(save_nodes.at(i));

          } catch(...) {}

          save_nodes.clear();
          if(!is_archive) output_node_srcml(*node, writer);
          is_archive = true;

        }

        if(node->extra) {
          
          freeNode(node);
          if(xmlTextReaderRead(reader) != 1) {done = true;   xmlTextWriterEndDocument(writer); return 1; }
          node = getNode(reader);
          break;
        }

        read_unit_start = true;
      }


      if(node->type == (xmlElementType)XML_READER_TYPE_END_ELEMENT) {

        break;
      }

    }

    if(!save_nodes.empty() && node->type == (xmlElementType)XML_READER_TYPE_ELEMENT
       && strcmp((const char *)node->name, "unit") != 0) {

      is_archive = true;
      is_single = true;

      try {

        for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size(); ++i)
          output_node_srcml(*save_nodes.at(i), writer);

        for(std::vector<xmlNodePtr>::size_type i = 0; i < save_nodes.size() - 1; ++i)
          freeNode(save_nodes.at(i));

      } catch(...) {}

      save_nodes.clear();

    }

    if(xmlTextReaderRead(reader) != 1) {done = true; return 0; }
    if(is_archive) freeNode(node);
    node = getNode(reader);

  }

  //freeNode(node);

  xmlTextWriterEndDocument(writer);

  return 1;

}

/**
 * readsrcML
 *
 * Read the next unit from a srcML Archive
 * and return it as a std::string. Uses
 * readsrcML(xmlTextWriterPtr writer).
 *
 * @returns string on success and finished return a 0.
 */
std::string * srcMLReader::readsrcML() {

  xmlBufferPtr buffer = xmlBufferCreate();
  xmlTextWriterPtr writer = xmlNewTextWriterMemory(buffer, 0);
  int status = readsrcML(writer);

  if(!status) {

    xmlFreeTextWriter(writer);
    xmlBufferFree(buffer);
    return 0;

  }

  int length = buffer->use;

  while(length > 0 && buffer->content[length - 1] == '\n')
    --length;

  std::string * unit = 0;
  try {

    unit = new std::string((const char *)buffer->content, length);

  } catch(...) {}

  xmlFreeTextWriter(writer);
  xmlBufferFree(buffer);

  return unit;

}

/**
 * output_node_srcml
 * @param node an XML node
 * @param writer an XML text writer
 * @param is_root boolean indicating if parsing root unit
 *
 * Output node as srcML.
 */
void output_node_srcml(const xmlNode & node, xmlTextWriterPtr writer) {

  bool isemptyelement = false;

  switch (node.type) {
  case XML_READER_TYPE_ELEMENT:

    // record if this is an empty element since it will be erased by the attribute copying
    isemptyelement = node.extra & 0x1;

    // start the element
    {

      std::string s = "";
      if(node.ns && node.ns->prefix) {

        s += ((char*) node.ns->prefix);
        s += ":";
      }
      s += (char*) node.name;

      xmlTextWriterStartElement(writer, BAD_CAST s.c_str());

    }

    if(strcmp((const char *)node.name, "unit") == 0) {

      xmlNsPtr xmlns = node.nsDef;
      while(xmlns) {

        std::string ns = xmlns->href ? (const char *)xmlns->href : "";
        if(ns != SRCML_CPP_NS_URI) {

          xmlns = xmlns->next;
          continue;

        }

        std::string prefix = "xmlns";
        if(xmlns->prefix) {

          prefix += ":";
          prefix += (const char *)xmlns->prefix;

        }

        xmlTextWriterWriteAttribute(writer, (const xmlChar *)prefix.c_str(), (const xmlChar *)ns.c_str());

        xmlns = xmlns->next;
      }
    }

    // copy all the attributes
    {
      xmlAttrPtr attribute = node.properties;
      while (attribute) {

        std::string s;
        if(attribute->ns && attribute->ns->prefix) {
          s = (const char *)attribute->ns->prefix;
          s += ":";

        }
        s += (const char *)attribute->name;

        xmlTextWriterWriteAttribute(writer, (const xmlChar *)s.c_str(), (const xmlChar *)attribute->children->content);
        attribute = attribute->next;
      }
    }

    // end now if this is an empty element
    if (isemptyelement) {

      xmlTextWriterEndElement(writer);
    }

    break;

  case XML_READER_TYPE_END_ELEMENT:
    xmlTextWriterEndElement(writer);
    break;

  case XML_READER_TYPE_COMMENT:
    xmlTextWriterWriteComment(writer, (const xmlChar *)node.content);
    break;

  case XML_READER_TYPE_TEXT:
  case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:

    // output the UTF-8 buffer escaping the characters.  Note that the output encoding
    // is handled by libxml
    for (unsigned char* p = (unsigned char*) node.content; *p != 0; ++p) {
      if (*p == '&')
        xmlTextWriterWriteRawLen(writer, BAD_CAST (unsigned char*) "&amp;", 5);
      else if (*p == '<')
        xmlTextWriterWriteRawLen(writer, BAD_CAST (unsigned char*) "&lt;", 4);
      else if (*p == '>')
        xmlTextWriterWriteRawLen(writer, BAD_CAST (unsigned char*) "&gt;", 4);
      else
        xmlTextWriterWriteRawLen(writer, BAD_CAST (unsigned char*) p, 1);
    }
    break;

  default:
    break;
  }
}
